/*-
 * Copyright (C)2000..2014 @BABOLO http://www.babolo.ru/\n\
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2000..2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: tree.c,v 1.7 2014/10/03 19:02:29 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3

#include <sys/types.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <babolo/BLINflag.h>
#include "deleted.h"

#define RES_LOW(A)                                                                                    \
    do {                                                                                              \
        ifBLIN_QX1("Resources low in ###%03d", (A));                                                  \
        parm->conf->res_low = true;                                                                   \
        break;                                                                                        \
    } while(0)

#define DUMPT(C,T)                                                                                    \
    do {                                                                                              \
        localtime_r(&(T), &t);                                                                        \
        ifBLIN_QX4( C" %04d-%02d-%02d %02d:%02d:%02d"                                                 \
                  , 1900 + t.tm_year, t.tm_mon, t.tm_mday, t.tm_hour, t.tm_min, t.tm_sec              \
                  );                                                                                  \
    } while(0)

static void
debug_tree(d_entry *tree, int level) {
    d_entry *l;
    int i;

    for (l = tree; l; l = l->next) {
        for (i = 0; i < level; i++) fputc('|', stderr);
        if  (l) {
            debug_date(stderr, l->time);
            fprintf( stderr, " %"BLIN_X" next %"BLIN_X" high %"BLIN_X" %s\n"
                   , BLIN_I(l), BLIN_I(l->next), BLIN_I(l->high), l->name
                   );
        } else {
            fprintf( stderr, "-----------------\n");
        }
        if  (l->high) debug_tree(l->high, level + 1);
}   }

static void
free_node(d_entry *node) {
    if  (node && node->name) free(node->name);
    free(node);
}

void
free_tree(d_entry *tree) {
    d_entry *d;

    while (tree) {
        if  (tree->high && tree->next) {
            d = tree->next;
            if  (d->high && d->next) {
                tree->next = d->high;
                d->high = tree;
                tree = d;
            } else if (d->high) {
                tree->next = d->high;
                free_node(d);
            } else if (d->next) {
                tree->next = d->next;
                free_node(d);
            } else {
                tree->next = NULL;
                free_node(d);
            }
        } else if (tree->high) {
            d = tree;
            tree = tree->high;
            free_node(d);
        } else if (tree->next) {
            d = tree;
            tree = tree->next;
            free_node(d);
        } else {
            free_node(tree);
            tree = NULL;
}   }   }

static time_t
youngest(time_t *a, time_t b) {
    if  ((!!b) && (*a < b)) *a = b;
    return(*a);
}

static time_t
select_time(deleted_pid *conf, struct stat *sb, aim *mask) {
    time_t result = 0;
    bool defined;

#   define blin_internal_flags (conf->flags & BLIN_MASK)
    defined = false;
    if  (   (mask->mask & (USE_O | USE_G | USE_R))
        &&  (!(mask->mask & USE_O) || sb->st_uid == mask->own)
        &&  (!(mask->mask & USE_G) || sb->st_gid == mask->grp)
        &&  (!(mask->mask & USE_R) || sb->st_mode == mask->mod)
        ) {
        result = NOW;
        defined = true;
    }
    if  (mask->mask & USE_A) {
        result = sb->st_atime;
        defined = true;
    }
    if  (mask->mask & USE_C) {
        if  (!defined) {
            result = sb->st_ctime;
        } else if (result < sb->st_ctime) {
            result = sb->st_ctime;
        }
        defined = true;
    }
    if  (mask->mask & USE_M) {
        if  (!defined) {
            result = sb->st_mtime;
        } else if (result < sb->st_mtime) {
            result = sb->st_mtime;
        }
        defined = true;
    }
    if  (mask->mask & USE_B) {
        if  (!defined) {
            result = sb->st_birthtime;
        } else if (result < sb->st_birthtime) {
            result = sb->st_birthtime;
        }
        defined = true;
    }
    result += mask->ttl;
    if  (!defined) {
        result = 0;
        ifBLIN_QX0("time mask is empty");
    }
    return(result);
#   undef blin_internal_flags
}

static char *
high_name(const char *path, const char *next) {
    char *next_name;

    asprintf(&next_name, "%s%s%s", path, (*path && path[strlen(path) - 1] == '/') ? "" : "/", next);
    return(next_name);
}

static d_entry *
create_n(const char *name, time_t time) {
    d_entry *d;

    d = malloc(sizeof(d_entry));
    if  (d) {
        d->high = NULL;
        d->next = NULL;
        d->name = name ? strdup(name) : NULL;
        d->time = time;
    }
    return(d);
}

static void
insert_n(d_entry **list, d_entry *d) {
    d_entry **l;
    int i;

    for (i = 0, l = list; *l && ((*l)->time < d->time); l = &((*l)->next)) ++i;
    if  (i) FREE_TREE(d->high);
    if  (*l) FREE_TREE((*l)->high);
    d->next = *l;
    *l = d;
}

d_entry *
/***********************************************************************
 *                                                                     */
build_tree(deleted_pid *parm, const char *path) {
/*                                                                     *
 ***********************************************************************
 * Build partial tree
 *              high       high       high       high
 * tree -> node ----> node ----> node ----> node ----> node  < the oldest node in the tree
 *          |          |          |                     |
 *          | next     | next     | next                | next
 *          V          V          V                     V
 *         node       node       node                  node
 *          |          |          |                     |
 *          | next     | next     | next                | next
 *          V          V          V                     V
 *         node       node       node                  node
 *          |          |                                |
 *          | next     | next   sorted                  | next
 *          V          V        by time                 V
 *         node       node                             node
 *
 *        sorted     sorted                           sorted
 *        by time    by time                          by time - oldest is highest,
 *                                                              younger is lower,
 *                                                              bubbles go down (in sort)
 *
 * Next link points to the node in the same directory as parent.
 * Next linked list sorted by time.
 * High link points to subdirectory.
 * The only high linked line exists - the way to the
 * oldest node in the directory.
 * Time of the directory is oldest time of files and subdirectories
 * of this directory if it is not empty.
 **********************************************************************
 */
    d_entry **list, *d, *l = NULL;
    time_t now, unsleep, linktime;
    struct stat sb, sb1;
    struct dirent *dp;
    char *next_name;
    struct tm t;
    DIR *dirp;
    int df, f;
    size_t s;

#   define blin_internal_flags (parm->conf->flags & BLIN_MASK)
    ifBLIN_QX2("+ %s~", path);
    list = &l;
    if  (!!lstat(path, &sb)) {
        ifBLIN_QW0("lstat %s~", path);
        return(NULL);
    }
    if  (!S_ISDIR(sb.st_mode)) {
        ifBLIN_QX0("not a dir %s~", path);
        return(NULL);
    }
    if  (!(dirp = opendir(path))) {
        ifBLIN_QW0("opendir %s~", path);
        return(NULL);
    }
    if  (0 > (df = dirfd(dirp))) {
        ifBLIN_QW0("dirfd %s~", path);
        return(NULL);
    }
    if  (fstat(df, &sb1)) {
        ifBLIN_QW0("fstat %s~", path);
        return(NULL);
    }
    if  ((sb.st_dev != sb1.st_dev) || (sb.st_ino != sb1.st_ino)) {
        ifBLIN_QX0("subst %s~", path);
        return(NULL);
    }
    s = strlen(parm->pidname);
    while ((dp = readdir(dirp)) != NULL) {
        now = NOW;
        if  (   !strcmp(dp->d_name, ".")
            ||  !strcmp(dp->d_name, "..")
            ||  (   dp->d_namlen = s
                &&  !strcmp(path, "/")
                &&  !strncmp(dp->d_name, parm->pidname, s)
            )   ) {
            ifBLIN_QX5(", skip %s~", dp->d_name);
    ;       continue;
    ;   }
    ;   if  (!!fchdir(df)) {
    ;       ifBLIN_QW0("fchdir");
    ;       FREE_TREE(*list);
    ;       break;
    ;   }
    ;   if  (!!lstat(dp->d_name, &sb)) {
            ifBLIN_QX1(", lstat %s~", dp->d_name);
    ;       continue;
    ;   }
    ;   switch (sb.st_mode & S_IFMT) {
        case S_IFLNK:
    ;       if  (!!parm->conf->lmask.ttl) {
    ;       ;   linktime = select_time(parm->conf, &sb, &parm->conf->lmask);
    ;       ;   DUMPT("LNK 0   ", linktime);
    ;       ;   f = 0;
    ;       ;   if  (parm->conf->lmask.mask & (USE_P | USE_L)) {
    ;       ;       /* XXXX     readlink() */
    ;       ;       if  (!!stat(dp->d_name, &sb)) {
    ;       ;           ifBLIN_QX1("stat %s~", dp->d_name);
    ;       ;           f = -1;
    ;       ;       } else if (parm->conf->lmask.mask & USE_P) {
    ;       ;           if  (S_ISDIR(sb.st_mode)) {
    ;       ;               youngest(&linktime, select_time(parm->conf, &sb, &parm->conf->dmask));
    ;       ;               DUMPT("LNK DIR ", linktime);
    ;       ;           } else if (S_ISREG(sb.st_mode)) {
    ;       ;               youngest(&linktime, select_time(parm->conf, &sb, &parm->conf->fmask));
    ;       ;               DUMPT("LNK REG ", linktime);
    ;       ;           } else {
    ;       ;               linktime = now + parm->conf->lmask.ttl;
    ;       ;               DUMPT("LNK LNK ", linktime);
    ;       ;   }   }   }
    ;       ;   unsleep = linktime + parm->conf->lmask.ttl;
    ;       ;   DUMPT("LNK 1   ", unsleep);
    ;       ;   if  (!f && ((unsleep + BONUS_LIFE) <= now) && (parm->conf->lmask.mask & USE_L)) {
    ;       ;       if  (0 > (f = open(dp->d_name, O_RDONLY | O_EXLOCK | O_NONBLOCK))) {
    ;       ;           ifBLIN_QW1("open %s~", dp->d_name);
    ;       ;           if  ((errno == EMFILE) || (errno == ENFILE)) {
    ;       ;               RES_LOW(701);                                                       /*###*/
    ;       ;           } else {
    ;       ;               unsleep = now + parm->conf->lmask.ttl;
    ;       ;               DUMPT("LNK lock", unsleep);
    ;       ;           }
    ;       ;       } else {
    ;       ;           close(f);
    ;       ;   }   }
    ;       ;   if  ((unsleep + BONUS_LIFE) <= now) {
    ;       ;       if  (!!unlink(dp->d_name)) {
    ;       ;           ifBLIN_QW1("unlink LNK %s", dp->d_name);
    ;       ;           unsleep = now + parm->conf->lmask.ttl;
    ;       ;           DUMPT("LNK rm- ", unsleep);
    ;       ;       } else {
    ;       ;           unsleep = 0;
    ;       ;           ifBLIN_QX4("LNK rm+");
    ;       ;   }   }
    ;       ;   if  (!!unsleep) {
    ;       ;       if  (!(d = create_n(dp->d_name, unsleep))) RES_LOW(600);                    /*###*/
    ;       ;       insert_n(list, d);
    ;       ;   }
    ;       ;   ifBLIN_QX5(", symlink %s~", dp->d_name);
    ;       }
    ;       continue;
    ;   case S_IFREG:
    ;       if  (!!parm->conf->fmask.ttl) {
    ;       ;   linktime = select_time(parm->conf, &sb, &parm->conf->fmask);
    ;       ;   DUMPT("REG 0   ", linktime);
    ;       ;   unsleep = linktime + parm->conf->fmask.ttl;
    ;       ;   DUMPT("REG 1   ", unsleep);
    ;       ;   if  (((unsleep + BONUS_LIFE) <= now) && (parm->conf->fmask.mask & USE_L)) {
    ;       ;       if  (0 > (f = open(dp->d_name, O_RDONLY | O_EXLOCK | O_NONBLOCK))) {
    ;       ;           ifBLIN_QW1("open %s~", dp->d_name);
    ;       ;           if  ((errno == EMFILE) || (errno == ENFILE)) {
    ;       ;               RES_LOW(707);                                                       /*###*/
    ;       ;           } else {
    ;       ;               unsleep = now + parm->conf->fmask.ttl;
    ;       ;               DUMPT("REG lock", unsleep);
    ;       ;           }
    ;       ;       } else {
    ;       ;           if  (!!fstat(f, &sb1)) {
    ;       ;               ifBLIN_QW1("fstat %s~", dp->d_name);
    ;       ;               unsleep = now + parm->conf->fmask.ttl;
    ;       ;               DUMPT("REG stat", unsleep);
    ;       ;           } else if ((sb.st_dev != sb1.st_dev) || (sb.st_ino != sb1.st_ino)) {
    ;       ;               ifBLIN_QX1( "d%u<>%u i%u<>%u %s~"
                                      , sb.st_dev, sb1.st_dev, sb.st_ino, sb1.st_ino, dp->d_name
                                      );
    ;       ;               unsleep = now + parm->conf->fmask.ttl;
    ;       ;               DUMPT("REG diff", unsleep);
    ;       ;           }
    ;       ;           close(f);
    ;       ;   }   }
    ;       ;   if  ((unsleep + BONUS_LIFE) <= now) {
    ;       ;       if  (!!unlink(dp->d_name)) {
    ;       ;           ifBLIN_QW1("unlink REG %s", dp->d_name);
    ;       ;           unsleep = now + parm->conf->fmask.ttl;
    ;       ;           DUMPT("REG rm- ", unsleep);
    ;       ;       } else {
    ;       ;           unsleep = 0;
    ;       ;           ifBLIN_QX4("REG rm+");
    ;       ;   }   }
    ;       ;   if  (!!unsleep) {
    ;       ;       if  (!(d = create_n(dp->d_name, unsleep))) RES_LOW(602);                    /*###*/
    ;       ;       insert_n(list, d);
    ;       ;   }
    ;       ;   ifBLIN_QX5(", file %s~", dp->d_name);
    ;       }
    ;       continue;
    ;   case S_IFDIR:
    ;       f = 0;
    ;       unsleep = 0;
    ;       if  (!!parm->conf->dmask.ttl) {
    ;       ;   linktime = select_time(parm->conf, &sb, &parm->conf->dmask);
    ;       ;   DUMPT("DIR 0   ", linktime);
    ;       ;   unsleep = linktime + parm->conf->dmask.ttl;
    ;       ;   DUMPT("DIR 1   ", unsleep);
    ;       ;   if  (((unsleep + BONUS_LIFE) <= now) && (parm->conf->dmask.mask & USE_L)) {
    ;       ;       if  (0 > (f = open(dp->d_name, O_RDONLY | O_EXLOCK | O_NONBLOCK))) {
    ;       ;           ifBLIN_QW1("open %s~", dp->d_name);
    ;       ;           if  ((errno == EMFILE) || (errno == ENFILE)) {
    ;       ;               RES_LOW(707);                                                       /*###*/
    ;       ;           } else {
    ;       ;               unsleep = now + parm->conf->dmask.ttl;
    ;       ;               DUMPT("DIR lock", unsleep);
    ;       ;           }
    ;       ;       } else {
    ;       ;           if  (!!fstat(f, &sb1)) {
    ;       ;               ifBLIN_QW1("fstat %s~", dp->d_name);
    ;       ;               unsleep = now + parm->conf->dmask.ttl;
    ;       ;               DUMPT("DIR stat", unsleep);
    ;       ;           } else if ((sb.st_dev != sb1.st_dev) || (sb.st_ino != sb1.st_ino)) {
    ;       ;               ifBLIN_QX1( "d%u<>%u i%u<>%u %s~"
                                      , sb.st_dev, sb1.st_dev, sb.st_ino, sb1.st_ino, dp->d_name
                                      );
    ;       ;               unsleep = now + parm->conf->dmask.ttl;
    ;       ;               DUMPT("DIR diff", unsleep);
    ;       ;           }
    ;       ;           close(f);
    ;       ;   }   }
    ;       ;   if  ((unsleep + BONUS_LIFE) <= now) {
    ;       ;       if  (!!rmdir(dp->d_name)) {
    ;       ;           ifBLIN_QW1("rmdir %s~", dp->d_name);
    ;       ;           unsleep = now + parm->conf->dmask.ttl;
    ;       ;           DUMPT("DIR rm- ", unsleep);
    ;       ;       } else {
    ;       ;           f = -1;
    ;       ;           ifBLIN_QX4("DIR rm+");
    ;       }   }   }
    ;       if  (0 <= f) {
    ;           if  (!(d = create_n(dp->d_name, unsleep))) RES_LOW(603);                        /*###*/
    ;           if  (!(next_name = high_name(path, dp->d_name))) RES_LOW(604);                  /*###*/
    ;           d->high = build_tree(parm, next_name);
    ;           free(next_name);
    ;           if  (d->high) d->time = d->high->time;
    ;           insert_n(list, d);
    ;       }
    ;       ifBLIN_QX5(", dir %s~", dp->d_name);
    ;       continue;
    }   }
    closedir(dirp);
    if  (!!chdir(path)) {
        ifBLIN_QW0("chdir %s~", path);
    }
    ifBLIN_QO3 debug_tree(*list, 0);
    ifBLIN_QX2("- %"BLIN_X" %"BLIN_X" %s~", list, *list, path);
    return(*list);
#   undef blin_internal_flags
}
